! ***********************************************************************
!
!   Copyright (C) 2012  Bill Paxton
!
!   this file is part of mesa.
!
!   mesa is free software; you can redistribute it and/or modify
!   it under the terms of the gnu general library public license as published
!   by the free software foundation; either version 2 of the license, or
!   (at your option) any later version.
!
!   mesa is distributed in the hope that it will be useful, 
!   but without any warranty; without even the implied warranty of
!   merchantability or fitness for a particular purpose.  see the
!   gnu library general public license for more details.
!
!   you should have received a copy of the gnu library general public license
!   along with this software; if not, write to the free software
!   foundation, inc., 59 temple place, suite 330, boston, ma 02111-1307 usa
!
! ***********************************************************************
 
      module run_star_extras 

      use star_lib
      use star_def
      use const_def
      use const_def
      use chem_def
      use num_lib
      use binary_def
      
      implicit none
      
      integer :: time0, time1, clock_rate
      real(dp), parameter :: expected_runtime = 1 ! minutes

      integer, parameter :: restart_info_alloc = 1
      integer, parameter :: restart_info_get = 2
      integer, parameter :: restart_info_put = 3
      
      
      contains
      
      subroutine extras_controls(id, ierr)
         integer, intent(in) :: id
         integer, intent(out) :: ierr
         ierr = 0
      end subroutine extras_controls
      
      
      integer function extras_startup(id, restart, ierr)
         integer, intent(in) :: id
         logical, intent(in) :: restart
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         integer :: restart_time, prev_time_used
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         if (.not. restart) then
            call system_clock(time0,clock_rate)
            call alloc_restart_info(s)
         else
            call unpack_restart_info(s)
            call system_clock(restart_time,clock_rate)
            prev_time_used = time1 - time0
            time1 = restart_time
            time0 = time1 - prev_time_used
         end if
         extras_startup = keep_going
      end function extras_startup
      

      integer function extras_check_model(id, id_extra)
         integer, intent(in) :: id, id_extra
         extras_check_model = keep_going
      end function extras_check_model


      integer function how_many_extra_history_columns(id, id_extra)
         integer, intent(in) :: id, id_extra
         how_many_extra_history_columns = 1
      end function how_many_extra_history_columns
      
      
      subroutine data_for_extra_history_columns(id, id_extra, n, names, vals, ierr)
         integer, intent(in) :: id, id_extra, n
         character (len=maxlen_profile_column_name) :: names(n)
         real(dp) :: vals(n)
         integer, intent(out) :: ierr
         real(dp) :: dt
         ierr = 0
         if (n /= 1) then
            stop 'bad n for data_for_extra_history_columns'
         end if
         dt = dble(time1 - time0) / clock_rate / 60
         names(1) = 'runtime_minutes'
         vals(1) = dt
      end subroutine data_for_extra_history_columns

      
      integer function how_many_extra_profile_columns(id, id_extra)
         integer, intent(in) :: id, id_extra
         how_many_extra_profile_columns = 0
      end function how_many_extra_profile_columns
      
      
      subroutine data_for_extra_profile_columns(id, id_extra, n, nz, names, vals, ierr)
         integer, intent(in) :: id, id_extra, n, nz
         character (len=maxlen_profile_column_name) :: names(n)
         real(dp) :: vals(nz,n)
         integer, intent(out) :: ierr
         integer :: k
         ierr = 0
      end subroutine data_for_extra_profile_columns
      
      subroutine extras_after_evolve(id, id_extra, ierr)
         integer, intent(in) :: id, id_extra
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         real(dp) :: dt
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         dt = dble(time1 - time0) / clock_rate / 60
         if (dt > 10*expected_runtime) then
            write(*,'(/,a30,2f18.6,a,/)') '>>>>>>> EXCESSIVE runtime', &
               dt, expected_runtime, '   <<<<<<<<<  ERROR'
         else
            write(*,'(/,a50,2f18.6,99i10/)') 'runtime, retries, backups, steps', &
               dt, expected_runtime, s% num_retries, s% num_backups, s% model_number
         end if
      end subroutine extras_after_evolve
      
      
      ! routines for saving and restoring data so can do restarts

      
      subroutine alloc_restart_info(s)
         type (star_info), pointer :: s
         call move_restart_info(s,restart_info_alloc)
      end subroutine alloc_restart_info
      
      
      subroutine unpack_restart_info(s)
         type (star_info), pointer :: s
         call move_restart_info(s,restart_info_get)
      end subroutine unpack_restart_info
      
      
      subroutine store_restart_info(s)
         type (star_info), pointer :: s
         call move_restart_info(s,restart_info_put)
      end subroutine store_restart_info
      
      
      subroutine move_restart_info(s,op)
         type (star_info), pointer :: s
         integer, intent(in) :: op
         
         integer :: i, j, num_ints, num_dbls, ierr
         
         i = 0
         ! call move_int or move_flg 
         call move_int(time0)
         call move_int(time1)
         
         num_ints = i
         
         i = 0
         ! call move_dbl 
         
         num_dbls = i
         
         if (op /= restart_info_alloc) return
         if (num_ints == 0 .and. num_dbls == 0) return
         
         ierr = 0
         call star_alloc_extras(s% id, num_ints, num_dbls, ierr)
         if (ierr /= 0) then
            write(*,*) 'failed in star_alloc_extras'
            write(*,*) 'alloc_extras num_ints', num_ints
            write(*,*) 'alloc_extras num_dbls', num_dbls
            stop 1
         end if
         
         contains
         
         subroutine move_dbl(dbl)
            real(dp) :: dbl
            i = i+1
            select case (op)
            case (restart_info_get)
               dbl = s% extra_work(i)
            case (restart_info_put)
               s% extra_work(i) = dbl
            end select
         end subroutine move_dbl
         
         subroutine move_int(int)
            integer :: int
            include 'formats'
            i = i+1
            select case (op)
            case (restart_info_get)
               !write(*,3) 'restore int', i, s% extra_iwork(i)
               int = s% extra_iwork(i)
            case (restart_info_put)
               !write(*,3) 'save int', i, int
               s% extra_iwork(i) = int
            end select
         end subroutine move_int
         
         subroutine move_flg(flg)
            logical :: flg
            i = i+1
            select case (op)
            case (restart_info_get)
               flg = (s% extra_iwork(i) /= 0)
            case (restart_info_put)
               if (flg) then
                  s% extra_iwork(i) = 1
               else
                  s% extra_iwork(i) = 0
               end if
            end select
         end subroutine move_flg

      end subroutine move_restart_info
      
      ! returns either keep_going or terminate.
      ! note: cannot request retry or backup; extras_check_model can do that.
      integer function extras_finish_step(id, id_extra)
        use crlibm_lib
        integer, intent(in) :: id, id_extra
        integer :: ierr, k
        type (star_info), pointer :: s
        real(dp) :: time, mdot_timescale, xsurf
    
        ierr = 0
        call star_ptr(id, s, ierr)
        if (ierr /= 0) return
        extras_finish_step = keep_going
        !call store_extra_info(s)

     
         ! Output when star has burnt given fraction of the helium in the core 
         ! during core helium burning
         s% lxtra1 = .false. 
         s% lxtra2 = .false.
         s% lxtra3 = .false.
         s% lxtra4 = .false.
         s% lxtra5 = .false.
         s% lxtra6 = .false.
         s% lxtra7 = .false.
         s% lxtra8 = .false.
         s% lxtra9 = .false.
         if (s%xa(s%net_iso(ih1), s%nz) .LE. 1d-3) then !TAMS
             if (s%xa(s%net_iso(ihe4), s%nz) .LE. 0.1) then  ! Helium left in core 0.1
                s% lxtra1 = .true.
                if ((s%lxtra1_old .eqv. .false.) .AND. (s% lxtra1 .eqv. .true.)) then
                   call output(id, ierr) 
                end if
             else if (s%xa(s%net_iso(ihe4), s%nz) .LE. 0.2) then  ! Helium left in core 0.2
                s% lxtra2 = .true.
                if ((s%lxtra2_old .eqv. .false.) .AND. (s% lxtra2 .eqv. .true.)) then
                   call output(id, ierr)
                end if
             else if (s%xa(s%net_iso(ihe4), s%nz) .LE. 0.3) then  ! Helium left in core 0.3
                s% lxtra3 = .true.
                if ((s%lxtra3_old .eqv. .false.) .AND. (s% lxtra3 .eqv. .true.)) then
                   call output(id, ierr)
                end if
             else if (s%xa(s%net_iso(ihe4), s%nz) .LE. 0.4) then  ! Helium left in core 0.4
                s% lxtra4 = .true.
                if ((s%lxtra4_old .eqv. .false.) .AND. (s% lxtra4 .eqv. .true.)) then
                   call output(id, ierr)
                end if
             else if (s%xa(s%net_iso(ihe4), s%nz) .LE. 0.5) then  ! Helium left in core 0.5
                s% lxtra5 = .true.
                if ((s%lxtra5_old .eqv. .false.) .AND. (s% lxtra5 .eqv. .true.)) then
                   call output(id, ierr)
                end if
             else if (s%xa(s%net_iso(ihe4), s%nz) .LE. 0.6) then  ! Helium left in core 0.6
                s% lxtra6 = .true.
                if ((s%lxtra6_old .eqv. .false.) .AND. (s% lxtra6 .eqv. .true.)) then
                   call output(id, ierr)
                end if
             else if (s%xa(s%net_iso(ihe4), s%nz) .LE. 0.7) then  ! Helium left in core 0.7
                s% lxtra7 = .true.
                if ((s%lxtra7_old .eqv. .false.) .AND. (s% lxtra7 .eqv. .true.)) then
                   call output(id, ierr)
                end if
             else if (s%xa(s%net_iso(ihe4), s%nz) .LE. 0.8) then  ! Helium left in core 0.8
                s% lxtra8 = .true.
                if ((s%lxtra8_old .eqv. .false.) .AND. (s% lxtra8 .eqv. .true.)) then
                   call output(id, ierr)
                end if
             else if (s%xa(s%net_iso(ihe4), s%nz) .LE. 0.9) then  ! Helium left in core 0.9
                s% lxtra9 = .true.
                if ((s%lxtra9_old .eqv. .false.) .AND. (s% lxtra9 .eqv. .true.)) then
                   call output(id, ierr)
                end if
             end if
         end if

                  
       contains
         
         subroutine output(id, ierr)
           ! use star_utils, only: get_name_for_restart_file
           use crlibm_lib
           interface
              subroutine save_restart_info(iounit, id, ierr)
                integer, intent(in) :: iounit
                integer, intent(in) :: id
                integer, intent(out) :: ierr
              end subroutine save_restart_info
           end interface
           integer, intent(in) :: id
           integer, intent(out) :: ierr     
           
           character (len=256) :: filename, num_str, fstring, modname
           type (star_info), pointer :: s
           integer :: num_digits
           
           call get_star_ptr(id, s, ierr)
           if (ierr /= 0) return
           !num_str = 'photo_23h55min'          

           if ((s%lxtra1_old .eqv. .false.) .and. (s%lxtra1 .eqv. .true.)) then
              !num_str = 'photo_TAMS'
              ! call star_write_model(id, 'TAMS.mod' , ierr)
              s% need_to_update_history_now = .true.
              s% need_to_save_profiles_now = .true.
              s% save_profiles_model_priority = 100 !for core he fraction 0.1 during helium burning
           else if ((s%lxtra2_old .eqv. .false.) .and. (s%lxtra2 .eqv. .true.)) then
              s% need_to_update_history_now = .true.
              s% need_to_save_profiles_now = .true.
              s% save_profiles_model_priority = 100 !for core he fraction 0.2 during helium burning
           else if ((s%lxtra3_old .eqv. .false.) .and. (s%lxtra3 .eqv. .true.)) then
              s% need_to_update_history_now = .true.
              s% need_to_save_profiles_now = .true.
              s% save_profiles_model_priority = 100 !for core he fraction 0.3 during helium burning
           else if ((s%lxtra4_old .eqv. .false.) .and. (s%lxtra4 .eqv. .true.)) then
              s% need_to_update_history_now = .true.
              s% need_to_save_profiles_now = .true.
              s% save_profiles_model_priority = 100 !for core he fraction 0.4 during helium burning
           else if ((s%lxtra5_old .eqv. .false.) .and. (s%lxtra5 .eqv. .true.)) then
              s% need_to_update_history_now = .true.
              s% need_to_save_profiles_now = .true.
              s% save_profiles_model_priority = 100 !for core he fraction 0.5 during helium burning
           else if ((s%lxtra6_old .eqv. .false.) .and. (s%lxtra6 .eqv. .true.)) then
              s% need_to_update_history_now = .true.
              s% need_to_save_profiles_now = .true.
              s% save_profiles_model_priority = 100 !for core he fraction 0.6 during helium burning
           else if ((s%lxtra7_old .eqv. .false.) .and. (s%lxtra7 .eqv. .true.)) then
              s% need_to_update_history_now = .true.
              s% need_to_save_profiles_now = .true.
              s% save_profiles_model_priority = 100 !for core he fraction 0.7 during helium burning
           else if ((s%lxtra8_old .eqv. .false.) .and. (s%lxtra8 .eqv. .true.)) then
              s% need_to_update_history_now = .true.
              s% need_to_save_profiles_now = .true.
              s% save_profiles_model_priority = 100 !for core he fraction 0.8 during helium burning
           else if ((s%lxtra9_old .eqv. .false.) .and. (s%lxtra9 .eqv. .true.)) then
              s% need_to_update_history_now = .true.
              s% need_to_save_profiles_now = .true.
              s% save_profiles_model_priority = 100 !for core he fraction 0.9 during helium burning
           end if
           
           !filename = trim(s% photo_directory) // '/' // trim(num_str) 
           !call output_to_file(filename, id, ierr) 
           
         end subroutine output
         
         
         subroutine output_to_file(filename, id, ierr)
         use utils_lib, only:alloc_iounit, free_iounit
         character (len=*) :: filename
         integer, intent(in) :: id
         integer, intent(out) :: ierr
         
         integer :: iounit, k
         type (star_info), pointer :: s
         character(len=256) :: iomsg

         include 'formats'

         call get_star_ptr(id, s, ierr)
         if (ierr /= 0) return

         iounit = alloc_iounit(ierr); if (ierr /= 0) return
         open(iounit, file=trim(filename), action='write', &
         status='replace', iostat=ierr, iomsg=iomsg, form='unformatted')
         if (ierr == 0) then
            s% most_recent_photo_name = trim(filename)
           ! call output_star_model(s, iounit)
            close(iounit)
         else
            write(*,*) trim(iomsg)
         endif
         call free_iounit(iounit)
         
      end subroutine output_to_file


!         if (extras_finish_step == terminate) s% termination_code = t_extras_finish_step
      end function extras_finish_step
      


      end module run_star_extras
      
